// Initialise fluid field pointer lists
PtrList<combustionModels::rhoCombustionModel> chemistryFluid(regions.size());
PtrList<volScalarField> rhoFluid(regions.size());
PtrList<volVectorField> UFluid(regions.size());
PtrList<volScalarField> TFluid(regions.size());
PtrList<surfaceScalarField> phiFluid(regions.size());
PtrList<uniformDimensionedVectorField> gFluid(regions.size());
PtrList<uniformDimensionedScalarField> hRefFluid(regions.size());
PtrList<compressible::turbulenceModel> turbulence(regions.size());
PtrList<volScalarField> p_rghFluid(regions.size());
PtrList<volScalarField> ghFluid(regions.size());
PtrList<surfaceScalarField> ghfFluid(regions.size());
PtrList<radiation::radiationModel> radiation(regions.size());
PtrList<volScalarField> KFluid(regions.size());
PtrList<volScalarField> dpdtFluid(regions.size());
List<scalar> initialMassFluid(regions.size());
PtrList<fv::options> fluidFvOptions(regions.size());

PtrList<diffusionModel> diffusionCoeffFluid(regions.size());
PtrList<PtrList<membraneModel> > membranesFluid(regions.size());
PtrList<PtrList<volScalarField> > membranesSpeciesMassFluxFluid(regions.size());    
PtrList<volScalarField> membranesMassFluxFluid(regions.size());
PtrList<volScalarField> membranesEnthalpyFluxFluid(regions.size());

// Populate field pointer lists
forAll(regions, i)
{
    Info<< "*** Reading thermophysical properties for region "
        << regions[i].name() << nl << endl;

    Info<< "    Adding to chemistrys\n" << endl;

    chemistryFluid.set
    (
        i,
        combustionModels::rhoCombustionModel::New(regions[i]).ptr()
    );

    Info<< "\n    Adding to rhos\n" << endl;
    rhoFluid.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "rho",
                runTime.timeName(),
                regions[i],
                IOobject::NO_READ,
                IOobject::AUTO_WRITE
            ),
            chemistryFluid[i].thermo().rho()
        )
    );

    Info<< "    Adding to Us\n" << endl;
    UFluid.set
    (
        i,
        new volVectorField
        (
            IOobject
            (
                "U",
                runTime.timeName(),
                regions[i],
                IOobject::MUST_READ,
                IOobject::AUTO_WRITE
            ),
            regions[i]
        )
    );

    Info<< "    Adding to phis\n" << endl;
    phiFluid.set
    (
        i,
        new surfaceScalarField
        (
            IOobject
            (
                "phi",
                runTime.timeName(),
                regions[i],
                IOobject::READ_IF_PRESENT,
                IOobject::AUTO_WRITE
            ),
            linearInterpolate(rhoFluid[i]*UFluid[i])
                & regions[i].Sf()
        )
    );

    Info<< "    Adding to gs\n" << endl;
    gFluid.set
    (
        i,
        new uniformDimensionedVectorField
        (
            IOobject
            (
                "g",
                runTime.constant(),
                regions[i],
                IOobject::MUST_READ,
                IOobject::NO_WRITE
            )
        )
    );

    Info<< "    Adding to hRefs\n" << endl;
    hRefFluid.set
    (
        i,
        new uniformDimensionedScalarField
        (
            IOobject
            (
                "hRef",
                runTime.constant(),
                regions[i],
                IOobject::READ_IF_PRESENT,
                IOobject::NO_WRITE
            ),
            dimensionedScalar("hRef", dimLength, 0)
        )
    );

    dimensionedScalar ghRef
    (
        mag(gFluid[i].value()) > SMALL
      ? gFluid[i]
          & (cmptMag(gFluid[i].value())/mag(gFluid[i].value()))*hRefFluid[i]
      : dimensionedScalar("ghRef", gFluid[i].dimensions()*dimLength, 0)
    );

    Info<< "    Adding turbulence \n" << endl;
    turbulence.set
    (
        i,
        compressible::turbulenceModel::New
        (
            rhoFluid[i],
            UFluid[i],
            phiFluid[i],
            chemistryFluid[i].thermo()
        ).ptr()
    );

    Info<< "\n    Adding to ghs\n" << endl;
    ghFluid.set
    (
        i,
        new volScalarField("gh", (gFluid[i] & regions[i].C()) - ghRef)
    );

    Info<< "    Adding to ghfs\n" << endl;
    ghfFluid.set
    (
        i,
        new surfaceScalarField("ghf", (gFluid[i] & regions[i].Cf()) - ghRef)
    );

    p_rghFluid.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "p_rgh",
                runTime.timeName(),
                regions[i],
                IOobject::MUST_READ,
                IOobject::AUTO_WRITE
            ),
            regions[i]
        )
    );

    // Force p_rgh to be consistent with p
    p_rghFluid[i] = chemistryFluid[i].thermo().p() - rhoFluid[i]*ghFluid[i];

    regions[i].setFluxRequired(p_rghFluid[i].name());

    radiation.set
    (
        i,
        radiation::radiationModel::New(chemistryFluid[i].thermo().T())
    );

    initialMassFluid[i] = fvc::domainIntegrate(rhoFluid[i]).value();

    Info<< "\n    Adding to Ks\n" << endl;
    KFluid.set
    (
        i,
        new volScalarField
        (
            "K",
            0.5*magSqr(UFluid[i])
        )
    );

    Info<< "    Adding to dpdts\n" << endl;
    dpdtFluid.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "dpdt",
                runTime.timeName(),
                regions[i]
            ),
            regions[i],
            dimensionedScalar
            (
                "dpdt",
                chemistryFluid[i].thermo().p().dimensions()/dimTime,
                0
            )
        )
    );

    Info<< "    Adding fvOptions\n" << endl;
    fluidFvOptions.set
    (
        i,
        new fv::options(regions[i])
    );

    turbulence[i].validate();

    Info<< "    Adding to diffusionCoeffs\n" << endl;

    diffusionCoeffFluid.set
    (
        i, 
        diffusionModel::New
        (
            regions[i],
            chemistryFluid[i],
            turbulence[i]
        ).ptr()
    );

    IOdictionary membraneDict
    (
        IOobject
        (
            "membraneProperties",
            runTime.time().constant(),
            regions[i],
            IOobject::MUST_READ_IF_MODIFIED,
            IOobject::NO_WRITE
        )
    );

    wordList membranesList
    (
        membraneDict.lookup("membranes")
    );

    Info<< "    Adding to membranes\n" << endl;

    if(membranesList.size() == 0)
        Info << "Warning: No membrane is defined!" << nl << endl;

    membranesFluid.set
    (
        i, 
        new PtrList<membraneModel>(membranesList.size())
    );

    PtrList<membraneModel>& membranes = membranesFluid[i];

    forAll(membranes, j)
    {
        membranes.set
        (
            j,
            membraneModel::New
            (
                i,
                regions,
                chemistryFluid,
                membranesList[j]
            ).ptr()
        );
    }

    Info<< "    Adding to membranesSpeciesMassFluxes\n" << endl;

    PtrList<volScalarField>& species = chemistryFluid[i].thermo().composition().Y();

    membranesSpeciesMassFluxFluid.set
    (
        i, 
        new PtrList<volScalarField>(species.size())
    );

    PtrList<volScalarField>& membranesSpeciesMassFluxes = membranesSpeciesMassFluxFluid[i];

    bool membraneIOOption(false);
    if(membranesList.size() > 0)
        membraneIOOption = membranesFluid[i][0].IOoption(regions[i]);

    if(membraneIOOption)
    {
        forAll (membranesSpeciesMassFluxes, j)
        {
            membranesSpeciesMassFluxes.set
            (
                j, 
                new volScalarField
                (
                    IOobject
                    (
                        IOobject::groupName("S", species[j].name()),
                        runTime.timeName(),
                        regions[i],
                        IOobject::NO_READ,
                        IOobject::AUTO_WRITE
                    ),
                    regions[i],
                    dimensionedScalar(species[j].name(), dimMass/dimVol/dimTime, 0.0)
                )
            );
        }  
    }
    else
    {
        forAll (membranesSpeciesMassFluxes, j)
        {
            membranesSpeciesMassFluxes.set
            (
                j, 
                new volScalarField
                (
                    IOobject
                    (
                        IOobject::groupName("S", species[j].name()),
                        runTime.timeName(),
                        regions[i],
                        IOobject::NO_READ,
                        IOobject::NO_WRITE
                    ),
                    regions[i],
                    dimensionedScalar(species[j].name(), dimMass/dimVol/dimTime, 0.0)
                )
            );
        }  
    }  

    Info<< "    Adding to membranesMassFluxes\n" << endl;
    membranesMassFluxFluid.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "membranesMassFlux",
                runTime.timeName(),
                regions[i]
            ),
            regions[i],
            dimensionedScalar
            (
                "membranesMassFlux", 
                dimMass/dimVol/dimTime, 
                0.0
            )
        )
    );

    Info<< "    Adding to membranesEnthalpyFluxes\n" << endl;
    membranesEnthalpyFluxFluid.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "membranesEnthalpyFlux",
                runTime.timeName(),
                regions[i]
            ),
            regions[i],
            dimensionedScalar
            (
                "membranesEnthalpyFlux", 
                dimEnergy/dimVol/dimTime, 
                0.0
            )
        )
    );
}