/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Namespace
    Foam::diffusionModels

Description
    Base class for calculating diffusion coefficients of fluids species

Class
    Foam::diffusionModel

SourceFiles
    diffusionModel.C

\*---------------------------------------------------------------------------*/

#ifndef diffusionModel_H
#define diffusionModel_H

#include "IOdictionary.H"
#include "typeInfo.H"
#include "volFields.H"
#include "dimensionedScalar.H"
#include "Pair.H"
#include "nearWallDist.H"
#include "autoPtr.H"
#include "runTimeSelectionTables.H"
#include "fvCFD.H"
#include "rhoCombustionModel.H"
#include "turbulentFluidThermoModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                           Class diffusionModel Declaration
\*---------------------------------------------------------------------------*/

class diffusionModel
{

protected:

    // Protected data
    const fvMesh& mesh_;
    const combustionModels::rhoCombustionModel& chemistry_;
    const compressible::turbulenceModel& turbulence_;


private:

    // Private Member Functions

        //- Disallow default bitwise copy construct
        diffusionModel(const diffusionModel&);

        //- Disallow default bitwise assignment
        void operator=(const diffusionModel&);


public:

    //- Runtime type information
    TypeName("diffusionModel");


    // Declare run-time constructor selection table

        declareRunTimeSelectionTable
        (
            autoPtr,
            diffusionModel,
            subModel,
            (
                const fvMesh& mesh,
                const combustionModels::rhoCombustionModel& chemistry,
                const compressible::turbulenceModel& turbulence
            ),
            (mesh, chemistry, turbulence)
        );


    // Constructors

        //- Construct from components
        diffusionModel
        (
            const word& type,
            const fvMesh& mesh,
            const combustionModels::rhoCombustionModel& chemistry,
            const compressible::turbulenceModel& turbulence
        );


    // Selectors

        //- Return a reference to the selected diffusion model
        static autoPtr<diffusionModel> New
        (
            const fvMesh& mesh,
            const combustionModels::rhoCombustionModel& chemistry,
            const compressible::turbulenceModel& turbulence
        );


    //- Destructor
    virtual ~diffusionModel()
    {}


    // Member Functions

        //- Return the effective diffusion per specie 
        virtual volScalarField effDiff(const word& specie) = 0;

        //- Read dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
