/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2015 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "membraneIO.H"
#include "fvMesh.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(membraneIO, 0);
}

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::membraneIO::membraneIO
(
    const fvMesh& mesh,
    const word& patchName
)
:
    mesh_(mesh),
    patchName_(patchName)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::Pair<Foam::dictionary> Foam::membraneIO::patchDicts()
{
    const word nbrPatchName_(nbrPatchName(mesh_, patchName_));

    return Pair<dictionary>
    (
        ownDict(mesh_).subDict(patchName_),
        nbrDict(mesh_, patchName_).subDict(nbrPatchName_)
    );
}

Foam::Pair<Foam::dictionary> Foam::membraneIO::subDicts
(
    const Pair<Foam::dictionary>& subdicts,
    const word subDictName
)
{
    dictionary ownSubDict_;
    dictionary nbrSubDict_;

    bool ownDictExist(false);
    bool nbrDictExist(false);

    if(subdicts.first().isDict(subDictName))
    {
        ownSubDict_ = subdicts.first().subDict(subDictName);
        ownDictExist = true; 
    }

    if(subdicts.second().isDict(subDictName))
    {
        nbrSubDict_ = subdicts.second().subDict(subDictName);
        nbrDictExist = true; 
    }

    if(!ownDictExist && !nbrDictExist)
    {
        readDictError(subdicts, subDictName);
    }
    if(!ownDictExist)
    {
        ownSubDict_ = nbrSubDict_;
    }
    if(!nbrDictExist)
    {
        nbrSubDict_ = ownSubDict_;
    }

    return Pair<dictionary>
    (
        ownSubDict_,
        nbrSubDict_
    );
}


Foam::scalar Foam::membraneIO::readScalar
(
    const Pair<dictionary>& subdicts, 
    const word prop
)
{
    scalar myDictValue(0);
    scalar nbrDictValue(0);

    bool myDictExist = subdicts.first().readIfPresent(prop, myDictValue);
    bool nbrDictExist = subdicts.second().readIfPresent(prop, nbrDictValue);

    if (!myDictExist && !nbrDictExist)
    {
        readDictError(subdicts, prop);
    }

    if (myDictExist && nbrDictExist)
    {
        if(myDictValue != nbrDictValue)
        {
            inconsitentValueError(subdicts, prop);
        }
    }

    if(myDictExist)
    {
        return myDictValue;
    }
    else 
    {
        return nbrDictValue;
    }
}


Foam::word Foam::membraneIO::readWord
(
    const Pair<dictionary>& subdicts, 
    const word prop
)
{
    word myDictValue("noWord");
    word nbrDictValue("noWord");

    bool myDictExist = subdicts.first().readIfPresent(prop, myDictValue);
    bool nbrDictExist = subdicts.second().readIfPresent(prop, nbrDictValue);

    if (!myDictExist && !nbrDictExist)
    {
        readDictError(subdicts, prop);
    }

    if (myDictExist && nbrDictExist)
    {
        if(myDictValue != nbrDictValue)
        {
            inconsitentValueError(subdicts, prop);
        }
    }

    if(myDictExist)
    {
        return myDictValue;
    }
    else 
    {
        return nbrDictValue;
    }
}


Foam::wordList Foam::membraneIO::readWordList
(
    const Pair<dictionary>& subdicts, 
    const word prop
)
{
    wordList myDictValue;
    wordList nbrDictValue;

    bool myDictExist = subdicts.first().readIfPresent(prop, myDictValue);
    bool nbrDictExist = subdicts.second().readIfPresent(prop, nbrDictValue);

    if (!myDictExist && !nbrDictExist)
    {
        readDictError(subdicts, prop);
    }

    bool valueExist(false);

    if (myDictExist && nbrDictExist)
    {
        forAll(myDictValue, myValue)
        {
            forAll(nbrDictValue, nbrValue)
            {
                if(myDictValue[myValue] == nbrDictValue[nbrValue])
                {
                    valueExist = true;
                }
            }
        }

        if(myDictValue.size() != nbrDictValue.size())
            valueExist = false;

        if(myDictValue.size() == 0 && nbrDictValue.size() == 0)
            valueExist = true;

        if(!valueExist)
        {
            inconsitentValueError(subdicts, prop);
        }
    }

    if(myDictExist)
    {
        return myDictValue;
    }
    else 
    {
        return nbrDictValue;
    }
}


Foam::Polynomial<8> Foam::membraneIO::readPolynom
(
    const Pair<dictionary>& subdicts, 
    const word prop
)
{
    Polynomial<8> myDictValue;
    Polynomial<8> nbrDictValue;

    bool myDictExist = subdicts.first().readIfPresent(prop+"<8>", myDictValue);
    bool nbrDictExist = subdicts.second().readIfPresent(prop+"<8>", nbrDictValue);

    if (!myDictExist && !nbrDictExist)
    {
        readDictError(subdicts, prop+"<8>");
    }

    if (myDictExist && nbrDictExist)
    {
        if(myDictValue != nbrDictValue)
        {
            inconsitentValueError(subdicts, prop+"<8>");
        }
    }

    if(myDictExist)
    {
        return myDictValue;
    }
    else 
    {
        return nbrDictValue;
    }
}


void Foam::membraneIO::readDictError
(
    const Pair<dictionary>& subdicts,
    const word prop 
)
{
    if(subdicts.first().name() == subdicts.second().name())
    {
        FatalErrorInFunction
            << "keyword " << prop 
            << " is undefined in dictionary "
            << subdicts.first().name() << nl
            << exit(FatalError);
    }
    else
    {
        FatalErrorInFunction
            << "keyword " << prop 
            << " should be at least defined in one of the following dictionaries:" << nl
            << subdicts.first().name() << nl
            << subdicts.second().name() << nl
            << exit(FatalError);
    }
}


void Foam::membraneIO::inconsitentValueError
(
    const Pair<dictionary>& subdicts,
    const word prop 
)
{
    FatalErrorInFunction
        << "keyword " << prop 
        << " should be the same in both dictionaries (or just define it in one dictionary):" << nl
        << subdicts.first().name() << nl
        << subdicts.second().name() << nl
        << exit(FatalError);
}


Foam::dictionary Foam::membraneIO::ownDict(const fvMesh& mesh)
{
    return  IOdictionary
            (
                IOobject
                (
                    "membraneProperties",
                    mesh.time().constant(),
                    mesh,
                    IOobject::MUST_READ_IF_MODIFIED,
                    IOobject::NO_WRITE,
                    false
                )
            );
}


Foam::dictionary Foam::membraneIO::nbrDict(const fvMesh& mesh, const word& patchName)
{
    label patchID = mesh.boundary().findPatchID(patchName);    

    // Get the coupling information from the mappedPatchBase
    const mappedPatchBase& mpp = refCast<const mappedPatchBase>
    (
        mesh.boundary()[patchID].patch()
    );

    const polyMesh& nbrMesh = mpp.sampleMesh();

    return  IOdictionary
            (
                IOobject
                (
                    "membraneProperties",
                    nbrMesh.time().constant(),
                    nbrMesh,
                    IOobject::MUST_READ_IF_MODIFIED,
                    IOobject::NO_WRITE,
                    false
                )
            );
}


Foam::word Foam::membraneIO::nbrPatchName
(
    const fvMesh& mesh, 
    const word& patchName
)
{
    label patchID = mesh.boundary().findPatchID(patchName);

    const mappedPatchBase& mpp = refCast<const mappedPatchBase>
    (
        mesh.boundary()[patchID].patch()
    );

    return mpp.samplePolyPatch().name();
}

// ************************************************************************* //
