/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "gasPermeation.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace membraneModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(gasPermeation, 0);
addToRunTimeSelectionTable(membraneModel, gasPermeation, subModel);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

gasPermeation::gasPermeation
(
    const label& myRegion,
    const PtrList<fvMesh>& meshes,
    const PtrList<combustionModels::rhoCombustionModel>& chemistries,
    const word& patchName,
    const word& modelName
)
:
    membraneModel(modelName, myRegion, meshes, chemistries, patchName),
    mesh_(meshes[myRegion]),
    myRegion_(myRegion),
    meshes_(meshes),
    chemistry_(chemistries[myRegion]),
    chemistries_(chemistries),
    patchName_(patchName),
    permeanceModelPtr_(permeanceModel::New(mesh_, chemistry_, patchName_))
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

volScalarField gasPermeation::specieMassFlux(const label& speciei)
{
    volScalarField permeateFlow
    (
        IOobject
        (
            "permeateFlow",
            mesh_
        ),
        mesh_,
        dimensionedScalar("permeateFlow", dimMass/dimVol/dimTime, 0.0)
    );

    label patchID = mesh_.boundary().findPatchID(patchName_);
    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const label nbrRegion(nbrRegionID());
    const label nbrPatchID(nbrPatch(mesh_, patchID));
    const word nbrPatchName(meshes_[nbrRegion].boundary()[nbrPatchID].name());

    if(permeateSpecie(Yi.name()))
    {
        // Current region
        const volScalarField& p = Yi.db().lookupObject<volScalarField>("p");
        const volScalarField& T = Yi.db().lookupObject<volScalarField>("T");
        
        volScalarField Xi(massToMoleFrac(speciei, specieComposition));

        volScalarField prtlPressure
        (
            IOobject
            (
                "prtlPressure",
                mesh_
            ),
            p*Xi
        );

        // Neighbour region
        const label nbrSpecie(nbrSpecieID(Yi.name()));
        const basicSpecieMixture& nbrSpecieComposition 
            = chemistries_[nbrRegion].thermo().composition();
        const volScalarField& nbrYi = nbrSpecieComposition.Y(nbrSpecie);
        const volScalarField& nbrP = nbrYi.db().lookupObject<volScalarField>("p");
        
        volScalarField nbrXi(massToMoleFrac(nbrSpecie, nbrSpecieComposition));

        volScalarField nbrPrtlPressure
        (
            IOobject
            (
                "nbrPrtlPressure",
                meshes_[nbrRegion]
            ),
            nbrP*nbrXi
        );  

        // Mapping neighbour region partial pressure
        volScalarField mppNbrPrtlPressure = 0.0 * prtlPressure;

        const mappedPatchBase& mpp = refCast<const mappedPatchBase>
        (
            Yi.boundaryField()[patchID].patch().patch()
        );
        
        mppNbrPrtlPressure.boundaryFieldRef()[patchID] 
      == // check fvPatchField.H line 572!
        nbrYi.db().lookupObject<volScalarField>
        ("nbrPrtlPressure").boundaryField()[nbrPatchID];

        mpp.distribute(mppNbrPrtlPressure.boundaryFieldRef()[patchID]);

        volScalarField permeance
        (
            permeanceModelPtr_->permeance(Yi.name(), patchName_, nbrPatchName)
        );

        // Defining pointers to fields
        const fvPatchScalarField& pP = p.boundaryField()[patchID];
        const fvPatchScalarField& pT = T.boundaryField()[patchID];
        const scalarField& volume = mesh_.V();
        const scalarField& pArea = mesh_.magSf().boundaryField()[patchID];
        scalarField& pPrtlPressure = prtlPressure.boundaryFieldRef()[patchID];
        scalarField& pMppNbrPrtlPressure = mppNbrPrtlPressure.boundaryFieldRef()[patchID];
        scalarField& pPermeance = permeance.boundaryFieldRef()[patchID];

        // Reading model and reference values from dictionary
        scalar refTemp = readScalar(subDicts(patchDicts(), Yi.name()), "refTemp");
        scalar refPress = readScalar(subDicts(patchDicts(), Yi.name()), "refPress");

        const fvPatch& currPatch = mesh_.boundary()[patchID];

        forAll(currPatch, facei)
        {
            label faceCelli = currPatch.faceCells()[facei];

            //calculating trans-membrane flow(TMF) at STP [m3/s]
            permeateFlow[faceCelli] 
          =
          - pPermeance[facei] * (pPrtlPressure[facei] 
          - pMppNbrPrtlPressure[facei]) * pArea[facei];

            //converting TMF from STP to current conditions [m3/s]
            permeateFlow[faceCelli] 
         *= 
            refPress / pP[facei] * pT[facei] / refTemp;

            //converting TMF from volume flowrate[m3/s] to volumetric mass flowrate[kg/m3/s]
            permeateFlow[faceCelli] 
         *= 
            specieComposition.rho(speciei, pP[facei], pT[facei]) / volume[faceCelli];
        }
    }

    return permeateFlow;
}


volScalarField gasPermeation::specieEnthalpyFlux
(
    const label& speciei,
    const volScalarField& transMembraneFlux
)
{
    volScalarField enthalpySource
    (
        IOobject
        (
            "enthalpySource",
            mesh_
        ),
        mesh_,
        dimensionedScalar("enthalpySource", dimEnergy/dimVol/dimTime, 0.0)
    );

    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const volScalarField& p = Yi.db().lookupObject<volScalarField>("p");
    const volScalarField& T = Yi.db().lookupObject<volScalarField>("T");

    forAll(enthalpySource, celli)
    {
        enthalpySource[celli] 
      += 
        specieComposition.Hs(speciei, p[celli], T[celli]) * transMembraneFlux[celli];
    }
    
    return enthalpySource;
}


bool gasPermeation::read()
{
    return true;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace membraneModels
} // End namespace Foam

// ************************************************************************* //
