/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "membraneModel.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

defineTypeNameAndDebug(membraneModel, 0);
defineRunTimeSelectionTable(membraneModel, subModel);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

membraneModel::membraneModel
(
    const word& type,
    const label& myRegion,
    const PtrList<fvMesh>& meshes,
    const PtrList<combustionModels::rhoCombustionModel>& chemistries,
    const word& patchName
)
:
    membraneIO(meshes[myRegion], patchName),
    mesh_(meshes[myRegion]),
    meshes_(meshes),
    chemistry_(chemistries[myRegion]),
    chemistries_(chemistries),
    patchName_(patchName)
{}


// * * * * * * * * * * * * * * * * * Selectors * * * * * * * * * * * * * * * //

autoPtr<membraneModel> membraneModel::New
(
    const label& myRegion,
    const PtrList<fvMesh>& meshes,
    const PtrList<combustionModels::rhoCombustionModel>& chemistries,
    const word& patchName
)
{
    label patchID = meshes[myRegion].boundary().findPatchID(patchName);

    membraneModel membraneProp("type", myRegion, meshes, chemistries, patchName);

    membraneProp.patchCheck(meshes[myRegion], patchID, patchName);
    membraneProp.patchTypeCheck(meshes[myRegion], patchID, patchName); 

    // Get the coupling information from the mappedPatchBase
    const mappedPatchBase& mpp = refCast<const mappedPatchBase>
    (
        meshes[myRegion].boundary()[patchID].patch()
    );

    const polyMesh& nbrMesh = mpp.sampleMesh();
    const label samplePatchi = mpp.samplePolyPatch().index();
    const fvPatch& nbrPatch =
        refCast<const fvMesh>(nbrMesh).boundary()[samplePatchi];

    word modelType 
        = membraneProp.readWord(membraneProp.patchDicts(), "membraneType");

    Info << "Selecting membrane model " << modelType 
         << " for patch " << patchName << nl << endl;

    subModelConstructorTable::iterator cstrIter =
        subModelConstructorTablePtr_->find(modelType);

    if (cstrIter == subModelConstructorTablePtr_->end())
    {
        FatalErrorInFunction
            << "Unknown membrane model type "
            << modelType << nl << nl
            << "Valid membrane model types:" << endl
            << subModelConstructorTablePtr_->sortedToc()
            << exit(FatalError);
    }

    membraneProp.checkNbrMembrane(meshes[myRegion], nbrMesh, patchName, nbrPatch);

    return autoPtr<membraneModel>
    (
        cstrIter()(myRegion, meshes, chemistries, patchName)
    );
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

volScalarField membraneModel::specieMassFlux(const label& speciei)
{
    return this->specieMassFlux(speciei);
}


volScalarField membraneModel::specieEnthalpyFlux
(
    const label& speciei,
    const volScalarField& transMembraneFlux
)
{
    return this->specieEnthalpyFlux(speciei, transMembraneFlux);
}


bool membraneModel::IOoption(const fvMesh& mesh)
{
    return readBool(ownDict(mesh).lookup("writeTransMembraneFluxes"));
}


void membraneModel::patchCheck
(
    const fvMesh& mesh, 
    const label& patchID, 
    const word& patchName
)
{
    if (patchID == -1)
    {
        FatalErrorInFunction
            << "Regin " << mesh.name() << " don't have any patch named "
            << patchName << endl
            << exit(FatalError);
    }
}


void membraneModel::patchTypeCheck
(
    const fvMesh& mesh, 
    const label& patchID, 
    const word& patchName
)
{
    if (mesh.boundary()[patchID].type() != "mappedWall")
    {
        FatalErrorInFunction
            << "Patch " << patchName << " is not a mappedWall!" << endl
            << exit(FatalError);
    }
}


void membraneModel::checkNbrMembrane
(
    const fvMesh& mesh,
    const polyMesh& nbrMesh,
    const word& patchName,
    const fvPatch& nbrPatch
)
{
    const wordList nbrMembranes(nbrDict(mesh, patchName).lookup("membranes"));

    bool nbrMembraneExist(false);

    forAll(nbrMembranes, membranei)
    {
        if(nbrMembranes[membranei] == nbrPatch.name())
            nbrMembraneExist = true;
    }

    if (!nbrMembraneExist)
    {
        FatalErrorInFunction
            << "Neighbour patch is not defined as a membrane! " << nl  << endl
            << exit(FatalError);
    }
}


bool membraneModel::permeateSpecie
(
    const word& specieName
)
{
    wordList permeateList(readWordList(patchDicts(), "permeates"));

    forAll(permeateList, permeatei)
    {
        if(permeateList[permeatei] == specieName)
            return true;
    }

    return false;
}


volScalarField membraneModel::massToMoleFrac
(
    const label& speciei,
    const basicSpecieMixture& specieComposition
)
{
    const volScalarField& Yi = specieComposition.Y(speciei);
    volScalarField moleFraction(0.0*Yi);

    moleFraction = Yi/specieComposition.W(speciei)*specieComposition.W();

    return moleFraction;
}


label membraneModel::nbrRegionID()
{
    // Get the coupling information from the mappedPatchBase
    label patchID = mesh_.boundary().findPatchID(patchName_); 
    
    const mappedPatchBase& mpp = refCast<const mappedPatchBase>
    (
        mesh_.boundary()[patchID].patch()
    );

    label nbrRegionID_(-1);

    forAll(meshes_, regioni)
    {
        if(meshes_[regioni].name() == mpp.sampleMesh().name())
            nbrRegionID_ = regioni;
    }

    return nbrRegionID_;
}


label membraneModel::nbrPatch
(
    const fvMesh& mesh, 
    const label& patchID
)
{
    const mappedPatchBase& mpp = refCast<const mappedPatchBase>
    (
        mesh.boundary()[patchID].patch()
    );

    const label nbrPatchID_ = mpp.samplePolyPatch().index();

    return nbrPatchID_;
}


label membraneModel::nbrSpecieID
(
    const word& specieName
)
{
    const label nbrRegion(nbrRegionID());
    const basicSpecieMixture& nbrSpecieComposition 
        = chemistries_[nbrRegion].thermo().composition();

    label nbrSpecieID_(-1);

    forAll(nbrSpecieComposition.Y(), nbrSpeciei)
    {
        if(nbrSpecieComposition.Y()[nbrSpeciei].name() == specieName)
            nbrSpecieID_ = nbrSpeciei;
    }

    return nbrSpecieID_;
}


bool membraneModel::read()
{
    return true;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
