/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Namespace
    Foam::membraneModels

Description
    Namespace for calculating trans membrane fluxes

Class
    Foam::membraneModel

SourceFiles
    membraneModel.C

\*---------------------------------------------------------------------------*/

#ifndef membraneModel_H
#define membraneModel_H

#include "IOdictionary.H"
#include "typeInfo.H"
#include "volFields.H"
#include "dimensionedScalar.H"
#include "Pair.H"
#include "nearWallDist.H"
#include "autoPtr.H"
#include "runTimeSelectionTables.H"
#include "fvCFD.H"
#include "rhoCombustionModel.H"
#include "mappedPatchBase.H"
#include "membraneIO.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

class membraneIO;

/*---------------------------------------------------------------------------*\
                           Class membraneModel Declaration
\*---------------------------------------------------------------------------*/

class membraneModel
:
    public membraneIO
{

protected:

    // Protected data
    const fvMesh& mesh_;
    const PtrList<fvMesh>& meshes_;
    const combustionModels::rhoCombustionModel& chemistry_;
    const PtrList<combustionModels::rhoCombustionModel>& chemistries_;
    const word patchName_;


private:

        //- Disallow default bitwise copy construct
        membraneModel(const membraneModel&);

        //- Disallow default bitwise assignment
        void operator=(const membraneModel&);


public:

    //- Runtime type information
    TypeName("membraneModel");


    // Declare run-time constructor selection table

        declareRunTimeSelectionTable
        (
            autoPtr,
            membraneModel,
            subModel,
            (
                const label& myRegion,
                const PtrList<fvMesh>& meshes,
                const PtrList<combustionModels::rhoCombustionModel>& chemistries,
                const word& patchName
            ),
            (myRegion, meshes, chemistries, patchName)
        );


    // Constructors

        //- Construct from components
        membraneModel
        (
            const word& type,
            const label& myRegion,
            const PtrList<fvMesh>& meshes,
            const PtrList<combustionModels::rhoCombustionModel>& chemistries,
            const word& patchName
        );


    // Selectors

        //- Return a reference to the selected membrane model
        static autoPtr<membraneModel> New
        (
            const label& myRegion,
            const PtrList<fvMesh>& meshes,
            const PtrList<combustionModels::rhoCombustionModel>& chemistries,
            const word& patchName
        );


    //- Destructor
    virtual ~membraneModel()
    {}


    // Member Functions

        //- Return the trans-membrane flux per specie 
        virtual volScalarField specieMassFlux(const label& speciei);

        //- Return the trans-membrane enthalpy flux per specie 
        virtual volScalarField specieEnthalpyFlux
        (
            const label& speciei,
            const volScalarField& transMembraneFlux
        );

        //- Option for saving tranasmembrane fluxes
        virtual bool IOoption(const fvMesh& mesh);
        
        //- Checking for existance of patch
        void patchCheck
        (
            const fvMesh& mesh, 
            const label& patchID, 
            const word& patchName
        );

        //- Checking for patch to be a wallMapped
        void patchTypeCheck
        (
            const fvMesh& mesh, 
            const label& patchID, 
            const word& patchName
        );

        //- Checking for the other side of patch to be a membrane
        void checkNbrMembrane
        (
            const fvMesh& mesh,
            const polyMesh& nbrMesh,
            const word& patchName,
            const fvPatch& nbrPatch
        );

        //- Checking if permeates exist in the owner and neighbour region
        bool permeateSpecie
        (
            const word& specieName
        );

        //- Conversion from mass fraction to mole fraction 
        volScalarField massToMoleFrac
        (
            const label& speciei,
            const basicSpecieMixture& composition
        );

        //- Return neighbour region ID 
        label nbrRegionID();

        //- Return neighbour patch ID
        label nbrPatch(const fvMesh& mesh, const label& patchID);

        //- Return specie ID in the neighbour region
        label nbrSpecieID(const word& specieName);

        //- Read dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
