/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "osmosis.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace membraneModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(osmosis, 0);
addToRunTimeSelectionTable(membraneModel, osmosis, subModel);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

osmosis::osmosis
(
    const label& myRegion,
    const PtrList<fvMesh>& meshes,
    const PtrList<combustionModels::rhoCombustionModel>& chemistries,
    const word& patchName,
    const word& modelName
)
:
    membraneModel(modelName, myRegion, meshes, chemistries, patchName),
    mesh_(meshes[myRegion]),
    myRegion_(myRegion),
    meshes_(meshes),
    chemistry_(chemistries[myRegion]),
    chemistries_(chemistries),
    patchName_(patchName),
    activityModelPtr_(activityModel::New(mesh_, chemistry_)),
    satPressModelPtr_(saturationPressureModel::New(mesh_, chemistry_)),
    permeanceModelPtr_(permeanceModel::New(mesh_, chemistry_, patchName_))
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

volScalarField osmosis::specieMassFlux(const label& speciei)
{
    label patchID = mesh_.boundary().findPatchID(patchName_);
    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const label nbrRegion(nbrRegionID());
    const label nbrPatchID(nbrPatch(mesh_, patchID));
    const word nbrPatchName(meshes_[nbrRegion].boundary()[nbrPatchID].name());

    const word myPhase(chemistry_.thermo().subDict("thermoType").lookup("regionPhase"));
    const word nbrPhase
                (
                    chemistries_[nbrRegion].thermo()
                    .subDict("thermoType").lookup("regionPhase")
                );

    word explicitSpecie
    (
        chemistry_.thermo().lookup("explicitSpecie")
    );

    wordList permeateList(readWordList(patchDicts(), "permeates"));
    wordList soluteList;

    forAll(permeateList, permeate)
    {
        if(permeateList[permeate]!=explicitSpecie) 
        {
            soluteList.append(permeateList[permeate]);
        }
    }


    regionsPhaseCheck(myPhase, nbrPhase);

    volScalarField permeateFlow
    (
        IOobject
        (
            "permeateFlow",
            mesh_
        ),
        mesh_,
        dimensionedScalar("permeateFlow", dimMass/dimVol/dimTime, 0.0)
    );

    permeateFlow = Yi.db().lookupObject<volScalarField>("S."+Yi.name());

    if(permeateSpecie(Yi.name()))
    {
        // Current region      
        volScalarField Xi(massToMoleFrac(speciei, specieComposition));

        volScalarField drivingForce
        (
            IOobject
            (
                "drivingForce",
                mesh_
            ),
            mesh_,
            dimensionedScalar("drivingForce", dimPressure, 0.0)
        );

        volScalarField osmoticPressure
        (
            IOobject
            (
                IOobject::groupName("osmoticPressure", Yi.name()),
                Yi.db().time().timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::AUTO_WRITE
            ),
            mesh_,
            dimensionedScalar("osmoticPressure", dimPressure, 0.0)
        );

        volScalarField osmoticPressureSolution
        (
            IOobject
            (
                "osmoticPressureSolution",
                Yi.db().time().timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::AUTO_WRITE
            ),
            mesh_,
            dimensionedScalar("osmoticPressureSolution", dimPressure, 0.0)
        );


        if(Yi.name() == explicitSpecie) //solvent
        {

            forAll(specieComposition.Y(), specie)
            {
                //Info << specieComposition.Y(solute).name() << endl;
                const volScalarField& Yspecie = specieComposition.Y(specie);
                drivingForce -=
                activityModelPtr_->activityCoeff(Yspecie.name())
              * satPressModelPtr_->saturationPressure(Yspecie.name());
            }

            osmoticPressureSolution = -1.0 * drivingForce;
            osmoticPressure += 
                activityModelPtr_->activityCoeff(Yi.name())
              * satPressModelPtr_->saturationPressure(Yi.name());

        }
        else //solute
        {
            drivingForce = Yi * dimensionedScalar("dimPressure", dimPressure, 1.0);
            osmoticPressure += 
                activityModelPtr_->activityCoeff(Yi.name())
              * satPressModelPtr_->saturationPressure(Yi.name());
        }

        if(Yi.db().time().outputTime())
        {
            osmoticPressure.write();
            if(Yi.name() == explicitSpecie) //solvent
            {
                osmoticPressureSolution.write();
            }

        }

        // Neighbour region
        const label nbrSpecie(nbrSpecieID(Yi.name()));
        const basicSpecieMixture& nbrSpecieComposition 
            = chemistries_[nbrRegion].thermo().composition();
        const volScalarField& nbrYi = nbrSpecieComposition.Y(nbrSpecie);
        
        volScalarField nbrXi(massToMoleFrac(nbrSpecie, nbrSpecieComposition));

        autoPtr<activityModel> nbrActivityModelPtr_
        (
            activityModel::runTimeNew
            (
                meshes_[nbrRegion], 
                chemistries_[nbrRegion]
            )
        );
        autoPtr<saturationPressureModel> nbrSatPressModelPtr_
        (
            saturationPressureModel::runTimeNew
            (
                meshes_[nbrRegion], 
                chemistries_[nbrRegion]
            )
        );

        volScalarField nbrDrivingForce
        (
            IOobject
            (
                "nbrDrivingForce",
                meshes_[nbrRegion]
            ),
            meshes_[nbrRegion],
            dimensionedScalar("nbrDrivingForce", dimPressure, 0.0)
        ); 

        if(Yi.name() == explicitSpecie) //solvent
        {
            forAll(nbrSpecieComposition.Y(), specie)
            {
                const volScalarField& nbrYspecie = nbrSpecieComposition.Y(specie);
                nbrDrivingForce -=
                nbrActivityModelPtr_->activityCoeff(nbrYspecie.name())
              * nbrSatPressModelPtr_->saturationPressure(nbrYspecie.name());
            }
        }
        else //solute
        {
            nbrDrivingForce = nbrYi * dimensionedScalar("dimPressure", dimPressure, 1.0);
        }

        // Mapping neighbour region partial pressure
        volScalarField mppNbrDrivingForce = 0.0 * drivingForce;

        const mappedPatchBase& mpp = refCast<const mappedPatchBase>
        (
            Yi.boundaryField()[patchID].patch().patch()
        );
        
        mppNbrDrivingForce.boundaryFieldRef()[patchID] 
      == // check fvPatchField.H line 572!
        nbrYi.db().lookupObject<volScalarField>
        ("nbrDrivingForce").boundaryField()[nbrPatchID];

        mpp.distribute(mppNbrDrivingForce.boundaryFieldRef()[patchID]);

        volScalarField permeance
        (
            permeanceModelPtr_->permeance(Yi.name(), patchName_, nbrPatchName)
        );

        // Defining pointers to fields
        const scalarField& volume = mesh_.V();
        const scalarField& pArea = mesh_.magSf().boundaryField()[patchID];
        scalarField& pDrivingForce = drivingForce.boundaryFieldRef()[patchID];
        scalarField& pMppNbrDrivingForce = mppNbrDrivingForce.boundaryFieldRef()[patchID];
        scalarField& pPermeance = permeance.boundaryFieldRef()[patchID];

        const fvPatch& currPatch = mesh_.boundary()[patchID];

        forAll(currPatch, facei)
        {
            label faceCelli = currPatch.faceCells()[facei];

            //calculating trans-membrane flow(TMF) [kg/s]
            permeateFlow[faceCelli] 
          =
          - pPermeance[facei] * (pDrivingForce[facei] 
          - pMppNbrDrivingForce[facei]) * pArea[facei];

            //converting to volumetric mass flow [Kg/m3/s]
            permeateFlow[faceCelli] 
         *= 
            1 / volume[faceCelli];
        }
    }

    return permeateFlow;
}


volScalarField osmosis::specieEnthalpyFlux
(
    const label& speciei,
    const volScalarField& transMembraneFlux
)
{
    const word myPhase(chemistry_.thermo().subDict("thermoType").lookup("regionPhase"));
    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const volScalarField& p = Yi.db().lookupObject<volScalarField>("p");
    const volScalarField& T = Yi.db().lookupObject<volScalarField>("T");

    dictionary thermoDict_;
    Polynomial<8> latentHeatTempCoeffs_;

    if(myPhase == "liquid")
    {
        thermoDict_ = IFstream
        (
            fileName(chemistry_.thermo().lookup("foamChemistryThermoFile")).expand()
        )();

        latentHeatTempCoeffs_
     =
        thermoDict_.subDict(Yi.name())
        .subDict("latentHeat").lookup("latentHeatTempCoeffs<" + Foam::name(8) + '>');
    }

    volScalarField enthalpySource
    (
        IOobject
        (
            "enthalpySource",
            mesh_
        ),
        mesh_,
        dimensionedScalar("enthalpySource", dimEnergy/dimVol/dimTime, 0.0)
    );

    forAll(enthalpySource, celli)
    {
        if(myPhase == "liquid")
        {   
            enthalpySource[celli] 
         += 
            (
                specieComposition.Hs(speciei, p[celli], T[celli]) 
              + latentHeatTempCoeffs_.value(T[celli])
            ) 
          * transMembraneFlux[celli];
        }
    }
    
    return enthalpySource;
}


void osmosis::regionsPhaseCheck
(
    const word& myPhase,
    const word& nbrPhase
)
{
    bool phaseCheck(false);

    if(myPhase == "liquid" && nbrPhase == "liquid")
    		phaseCheck = true;

    if(!phaseCheck)
    {
        FatalErrorInFunction
            << "Unknown phase types or wrong phase combination ("
            << myPhase << ", " << nbrPhase << ")" << nl << nl
            << "Valid phase types are:\n\n2\n(\nliquid\nliquid\n)" << nl << nl
            << "For osmosis both phases should be a liquid!" << endl
            << exit(FatalError);
    }
}


bool osmosis::read()
{
    return true;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace membraneModels
} // End namespace Foam

// ************************************************************************* //
