/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "pervaporation.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace membraneModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(pervaporation, 0);
addToRunTimeSelectionTable(membraneModel, pervaporation, subModel);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

pervaporation::pervaporation
(
    const label& myRegion,
    const PtrList<fvMesh>& meshes,
    const PtrList<combustionModels::rhoCombustionModel>& chemistries,
    const word& patchName,
    const word& modelName
)
:
    membraneModel(modelName, myRegion, meshes, chemistries, patchName),
    mesh_(meshes[myRegion]),
    myRegion_(myRegion),
    meshes_(meshes),
    chemistry_(chemistries[myRegion]),
    chemistries_(chemistries),
    patchName_(patchName),
    activityModelPtr_(activityModel::New(mesh_, chemistry_)),
    satPressModelPtr_(saturationPressureModel::New(mesh_, chemistry_)),
    permeanceModelPtr_(permeanceModel::New(mesh_, chemistry_, patchName_))
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

volScalarField pervaporation::specieMassFlux(const label& speciei)
{
    label patchID = mesh_.boundary().findPatchID(patchName_);
    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const label nbrRegion(nbrRegionID());
    const label nbrPatchID(nbrPatch(mesh_, patchID));
    const word nbrPatchName(meshes_[nbrRegion].boundary()[nbrPatchID].name());

    const word myPhase(chemistry_.thermo().subDict("thermoType").lookup("regionPhase"));
    const word nbrPhase
                (
                    chemistries_[nbrRegion].thermo()
                    .subDict("thermoType").lookup("regionPhase")
                );

    regionsPhaseCheck(myPhase, nbrPhase);

    volScalarField permeateFlow
    (
        IOobject
        (
            "permeateFlow",
            mesh_
        ),
        mesh_,
        dimensionedScalar("permeateFlow", dimMass/dimVol/dimTime, 0.0)
    );

    if(permeateSpecie(Yi.name()))
    {
        // Current region
        const volScalarField& p = Yi.db().lookupObject<volScalarField>("p");
        
        volScalarField Xi(massToMoleFrac(speciei, specieComposition));

        volScalarField prtlPressure
        (
            IOobject
            (
                "prtlPressure",
                mesh_
            ),
            mesh_,
            dimensionedScalar("prtlPressure", dimPressure, 0.0)
        );

        if(myPhase == "gas")
        {
            prtlPressure = p*Xi;
        }
        else
        {
            prtlPressure = Xi
          * activityModelPtr_->activityCoeff(Yi.name())
          * satPressModelPtr_->saturationPressure(Yi.name());
        }

        // Neighbour region
        const label nbrSpecie(nbrSpecieID(Yi.name()));
        const basicSpecieMixture& nbrSpecieComposition 
            = chemistries_[nbrRegion].thermo().composition();
        const volScalarField& nbrYi = nbrSpecieComposition.Y(nbrSpecie);
        const volScalarField& nbrP = nbrYi.db().lookupObject<volScalarField>("p");
        
        volScalarField nbrXi(massToMoleFrac(nbrSpecie, nbrSpecieComposition));

        autoPtr<activityModel> nbrActivityModelPtr_
        (
            activityModel::runTimeNew
            (
                meshes_[nbrRegion], 
                chemistries_[nbrRegion]
            )
        );
        autoPtr<saturationPressureModel> nbrSatPressModelPtr_
        (
            saturationPressureModel::runTimeNew
            (
                meshes_[nbrRegion], 
                chemistries_[nbrRegion]
            )
        );

        volScalarField nbrPrtlPressure
        (
            IOobject
            (
                "nbrPrtlPressure",
                meshes_[nbrRegion]
            ),
            meshes_[nbrRegion],
            dimensionedScalar("nbrPrtlPressure", dimPressure, 0.0)
        ); 

        if(nbrPhase == "gas")
        {
            nbrPrtlPressure = nbrP*nbrXi;
        }
        else
        {
            nbrPrtlPressure = nbrXi
          * nbrActivityModelPtr_->activityCoeff(Yi.name())
          * nbrSatPressModelPtr_->saturationPressure(Yi.name());
        } 
        
        // Mapping neighbour region partial pressure
        volScalarField mppNbrPrtlPressure = 0.0 * prtlPressure;

        const mappedPatchBase& mpp = refCast<const mappedPatchBase>
        (
            Yi.boundaryField()[patchID].patch().patch()
        );
        
        mppNbrPrtlPressure.boundaryFieldRef()[patchID] 
      == // check fvPatchField.H line 572!
        nbrYi.db().lookupObject<volScalarField>
        ("nbrPrtlPressure").boundaryField()[nbrPatchID];

        mpp.distribute(mppNbrPrtlPressure.boundaryFieldRef()[patchID]);

        volScalarField permeance
        (
            permeanceModelPtr_->permeance(Yi.name(), patchName_, nbrPatchName)
        );

        // Defining pointers to fields
        const scalarField& volume = mesh_.V();
        const scalarField& pArea = mesh_.magSf().boundaryField()[patchID];
        scalarField& pPrtlPressure = prtlPressure.boundaryFieldRef()[patchID];
        scalarField& pMppNbrPrtlPressure = mppNbrPrtlPressure.boundaryFieldRef()[patchID];
        scalarField& pPermeance = permeance.boundaryFieldRef()[patchID];

        const fvPatch& currPatch = mesh_.boundary()[patchID];

        forAll(currPatch, facei)
        {
            label faceCelli = currPatch.faceCells()[facei];

            //calculating molar trans-membrane flow(TMF) [mol/s]
            permeateFlow[faceCelli] 
          =
          - pPermeance[facei] * (pPrtlPressure[facei] 
          - pMppNbrPrtlPressure[facei]) * pArea[facei];

            //converting from molar flow to volumetric mass flow [Kg/m3/s]
            permeateFlow[faceCelli] 
         *= 
            specieComposition.W(speciei) / volume[faceCelli] / 1000.0;
        }
    }

    return permeateFlow;
}


volScalarField pervaporation::specieEnthalpyFlux
(
    const label& speciei,
    const volScalarField& transMembraneFlux
)
{
    const word myPhase(chemistry_.thermo().subDict("thermoType").lookup("regionPhase"));
    const basicSpecieMixture& specieComposition = chemistry_.thermo().composition();
    const volScalarField& Yi = specieComposition.Y(speciei);
    const volScalarField& p = Yi.db().lookupObject<volScalarField>("p");
    const volScalarField& T = Yi.db().lookupObject<volScalarField>("T");

    dictionary thermoDict_;
    Polynomial<8> latentHeatTempCoeffs_;

    if(myPhase == "liqiud")
    {
        thermoDict_ = IFstream
        (
            fileName(chemistry_.thermo().lookup("foamChemistryThermoFile")).expand()
        )();

        latentHeatTempCoeffs_
     =
        thermoDict_.subDict(Yi.name())
        .subDict("latentHeat").lookup("latentHeatTempCoeffs<" + Foam::name(8) + '>');
    }

    volScalarField enthalpySource
    (
        IOobject
        (
            "enthalpySource",
            mesh_
        ),
        mesh_,
        dimensionedScalar("enthalpySource", dimEnergy/dimVol/dimTime, 0.0)
    );

    forAll(enthalpySource, celli)
    {
        if(myPhase == "gas")
        {   
            enthalpySource[celli] 
         += 
            specieComposition.Hs(speciei, p[celli], T[celli]) * transMembraneFlux[celli];
        }
        else
        {
            enthalpySource[celli] 
         += 
            (
                specieComposition.Hs(speciei, p[celli], T[celli]) 
              + latentHeatTempCoeffs_.value(T[celli])
            ) 
          * transMembraneFlux[celli];
        }
    }
    
    return enthalpySource;
}


void pervaporation::regionsPhaseCheck
(
    const word& myPhase,
    const word& nbrPhase
)
{
    bool phaseCheck(false);

    if(myPhase == "gas" && nbrPhase == "liquid")
        phaseCheck = true;

    if(myPhase == "liquid" && nbrPhase == "gas")
        phaseCheck = true;

    if(!phaseCheck)
    {
        FatalErrorInFunction
            << "Unknown phase types or wrong phase combination ("
            << myPhase << ", " << nbrPhase << ")" << nl << nl
            << "Valid phase types are:\n\n2\n(\ngas\nliquid\n)" << nl << nl
            << "For pervaporation one phase should be liquid and one should be a gas!" << endl
            << exit(FatalError);
    }
}


bool pervaporation::read()
{
    return true;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace membraneModels
} // End namespace Foam

// ************************************************************************* //
