/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::membraneModels::pervaporation

Description
    Trans-membrane flux is calculated based on solution-diffusion mass transfer
    mechanism for pervaporation based units.

    References:
    \verbatim
        Rom, A., Miltner, A., Wukovits, W., & Friedl, A. (2016). 
        Energy saving potential of hybrid membrane and distillation process 
        in butanol purification: Experiments, modelling and simulation. 
        Chemical Engineering and Processing: Process Intensification, 104, 201-211.
    \endverbatim

    Sample dictionary:
    \verbatim
        permeates
        (
            CO2
        );

        membraneType    pervaporation;

        CO2
        {
            n                   0;          //[ 0  0  0  0  0  0  0 ]
            permeance           5.91e-10;   //[-1  2  1  0  0  0  0 ]
            activationEnergy    0;          //[ 1  2 -2  0 -1  0  0 ]
            refTemp             298.15;     
            refPress            1e5;
        }
    \endverbatim

SourceFiles
    pervaporation.C

\*---------------------------------------------------------------------------*/

#ifndef pervaporation_H
#define pervaporation_H

#include "membraneModel.H"
#include "activityModel.H"
#include "saturationPressureModel.H"
#include "permeanceModel.H"
#include "Polynomial.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace membraneModels
{

/*---------------------------------------------------------------------------*\
                           Class pervaporation Declaration
\*---------------------------------------------------------------------------*/

class pervaporation
:
    public membraneModel
{

protected:

    const fvMesh& mesh_;
    const label myRegion_;
    const PtrList<fvMesh>& meshes_;
    const combustionModels::rhoCombustionModel& chemistry_;
    const PtrList<combustionModels::rhoCombustionModel>& chemistries_;
    const word patchName_;

private:
        autoPtr<activityModel> activityModelPtr_;
        autoPtr<saturationPressureModel> satPressModelPtr_;
        autoPtr<permeanceModel> permeanceModelPtr_;
    
public:

    //- Runtime type information
    TypeName("pervaporation");

    // Constructors

        //- Construct from components
        pervaporation
        (
            const label& myRegion,
            const PtrList<fvMesh>& meshes,
            const PtrList<combustionModels::rhoCombustionModel>& chemistries,
            const word& patchName,
            const word& modelName = typeName
        );


    //- Destructor
    virtual ~pervaporation()
    {}


    // Member Functions

        //- Return the trans-membrane flux per specie 
        virtual volScalarField specieMassFlux(const label& speciei);

        //- Return the trans-membrane enthalpy flux per specie 
        virtual volScalarField specieEnthalpyFlux
        (
            const label& speciei,
            const volScalarField& transMembraneFlux
        );

        //- Check if one region is liquid and the other one is a gas
        void regionsPhaseCheck(const word& myPhase, const word& nbrPhase);

        //- Read dictionary
        virtual bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace membraneModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
